NOW = 'Now'
EP_VAR = '?E'
UTT_VAR = '?U'

MOOD_DECL = 'decl'
MOOD_QUES = 'ques'
MOOD_IMPER = 'imper'

SPEAKER = '^Speaker'
HEARER = '^Hearer'

EXISTS = 'exists'
AND = 'and.cc'
CH = '**'
MULT = 'mult'

SAME_TIME = 'same-time.p'
BEFORE = 'before.p'
AFTER = 'after.p'
AT_OR_BEFORE = 'at-or-before.p'
JUST_AFTER = 'just-after.p'
AT_ABOUT = 'at-about.p'
IMPINGES_ON = 'impinges-on.p'
ORIENTS = 'orients.p'

TELL = 'tell.v'
ASK = 'ask.v'
INSTRUCT = 'instruct.v'

THAT = 'that'
ANS_TO = 'ans-to'
KA = 'ka'

PAST = 'past'
PRES = 'pres'
FUTR = 'futr'
PERF = 'perf'